from typing import Optional

import chardet
from chardet.universaldetector import UniversalDetector

from .line_analysis_result import LineAnalysisResult

ALLOWED_ENCODINGS = (None, "ascii")


def get_encoding(*filenames: str) -> dict[str, chardet.ResultDict]:
    """Return encoding information for *filenames*.

    E.g.:

    {
        '__init__.py': {'encoding': 'ascii', 'confidence': 1.0, 'language': ''},
        '__main__.py': {'encoding': 'ascii', 'confidence': 1.0, 'language': ''},
        'cli.py': {'encoding': 'ascii', 'confidence': 1.0, 'language': ''}
        'utils.py': {'encoding': 'ascii', 'confidence': 1.0, 'language': ''}
    }

    """
    detector = UniversalDetector()
    encodings = {}
    for filename in filenames:
        detector.reset()
        with open(filename, "rb") as fh:
            for line in fh.readlines():
                detector.feed(line)
                if detector.done:
                    break
        detector.close()
        encodings[filename] = detector.result
    return encodings


def get_non_ascii_files(*filenames: str) -> dict[str, Optional[str]]:
    """Return dict of `filename: encoding` for files with not allowed encoding."""
    encoding_info = get_encoding(*filenames)
    return {
        filename: info["encoding"]
        for filename, info in encoding_info.items()
        if info["encoding"] not in ALLOWED_ENCODINGS
    }


def get_non_ascii_lines(filename: str) -> list[LineAnalysisResult]:
    """Return all lines with non-ASCII chars from *filename*."""

    results = []
    with open(filename) as f:
        for line_num, line in enumerate(f, start=1):
            result = LineAnalysisResult(line_num, line)

            for char_pos, char in enumerate(line, start=1):
                if not char.isascii():
                    result.add_char(char, char_pos)

            if result.chars:
                results.append(result)

    return results


def extract_context(line: str, center_position: int, context_width: int) -> str:
    """Return a context of the lines containing non-ASCII characters."""

    chars_before = context_width // 2
    chars_after = context_width // 2

    start_index = max(0, center_position - chars_before)
    end_index = min(len(line), center_position + chars_after)

    context = line[start_index:end_index]

    return context



from ski_lint.cli import get_args, run
from ski_lint.utils import get_encoding, get_non_ascii_files, get_non_ascii_lines

__all__ = ["get_args", "run", "get_encoding", "get_non_ascii_files", "get_non_ascii_lines"]



"""A pre-commit hook, that rejects files containing non-ASCII characters."""

import logging
import sys
from argparse import ArgumentParser, Namespace
from importlib.metadata import version

from .utils import extract_context, get_non_ascii_files, get_non_ascii_lines

logging.basicConfig(
    format="{asctime} - {levelname} - {message}",
    level=logging.INFO,
    style="{",
    stream=sys.stdout,
)
log = logging.getLogger(__name__)

description = __doc__.strip()


def get_args() -> Namespace:
    ap = ArgumentParser(description=description)
    ap.add_argument("--version", action="version", version=version("ski_lint"))
    ap.add_argument(
        "--check",
        action="store_true",
        help="return code is `1`, when non-ASCII files are found",
    )
    ap.add_argument("filenames", nargs="+", metavar="FILENAME", help="path to the files to check")
    ap.add_argument("--context-width", type=int, default=50, help="width of the context of the non-ASCII line")
    args = ap.parse_args()
    return args


def run(*filenames: str, check: bool = False, context_width: int) -> int:
    bad_encodings = get_non_ascii_files(*filenames)

    if bad_encodings:
        for filename, encoding in bad_encodings.items():
            non_ascii_lines = get_non_ascii_lines(filename)

            for line_result in non_ascii_lines:
                for char, char_positions in line_result.chars.items():
                    for char_pos in char_positions:
                        context = extract_context(line_result.line, char_pos, context_width)

                        error_msg = f"{filename} ({encoding}), "
                        error_msg += f"line {line_result.line_num}, "
                        error_msg += f"pos {char_pos}, "
                        if not char.isprintable():
                            error_msg += f"non-printable char U+{ord(char):0x}, "
                            context = context.replace(char, f"U+{ord(char):0x}")
                        else:
                            error_msg += f"char '{char}', "
                        error_msg += f"context: '{context}'"

                        log.error(error_msg)

    else:
        log.info("NON-ASCII CHECK: OK")

    if bad_encodings and check:
        return 1

    return 0


def main() -> None:
    args = get_args()
    sys.exit(run(*args.filenames, check=args.check, context_width=args.context_width))



class LineAnalysisResult:
    def __init__(self, line_num: int, line: str) -> None:
        self.line_num = line_num
        self.line = line.rstrip("\r\n")
        self.chars: dict[str, list[int]] = {}

    def add_char(self, char: str, char_pos: int) -> None:
        self.chars.setdefault(char, []).append(char_pos)
